#include "stdafx.h"

#include <stdio.h>

#include "loadbmp.h"
#include "math.h"

int LoadBitmapFromFile(char *name, CBitmap **bmp)
{
	//file handle used in all file operations
    FILE *in;
    LONG width,height;
    WORD bpp;
	int numColours;
	BITMAPINFO *bi;
	BITMAP bm;
	int ret_val;
	int i;

	if (strlen(name) <= 0)
		return BMP_LOAD_ERROR;

    //open the file for reading in binary mode    
    in=fopen(name,"rb");

	if (in == NULL)
	{
		return BMP_LOAD_ERROR;
	}

	BITMAPFILEHEADER bmfh;

	fread(&bmfh,sizeof(BITMAPFILEHEADER),1,in);

	BITMAPINFOHEADER bmih;

	fread(&bmih,sizeof(BITMAPINFOHEADER),1,in);	//set the number of colours
	
	numColours=1 << bmih.biBitCount;

	bi = (BITMAPINFO *) new char[sizeof(BITMAPINFO) + numColours * sizeof(RGBQUAD)];

	//load the palette for 8 bits per pixel
	if(bmih.biBitCount == 8) {
		fread(bi->bmiColors,sizeof(RGBQUAD),numColours,in);
	}

	BYTE *tempPixelData;
	int size;

	size=bmfh.bfSize-bmfh.bfOffBits;
	width = bmih.biWidth;
	height = bmih.biHeight;
	bpp = bmih.biBitCount;

	// check if allowed dimensions
	if (width != height || bpp != 8)
	{
		// possible texture
		for (i = 2; i < 20; i++)
		{
			if (pow(2, i) == width) break;
		}

		if (i == 20)
		{
			return BMP_DIMENSION_ERROR;
		}
		else
		{
			ret_val = BMP_TEXT;
		}
	}
	else
	{
		for (i = 2; i < 20; i++)
		{
			if (pow(2, i) + 1 == width) break;
		}

		if (i == 20)
		{
			for (i = 2; i < 20; i++)
			{
				if (pow(2, i) == width) break;
			}

			if (i == 20)
			{
				return BMP_DIMENSION_ERROR;
			}
			else
			{
				ret_val = BMP_TEXT;
			}
		}
		else
		{

			ret_val = BMP_MAP;
		}
	}

	tempPixelData=new BYTE[size];

	if(tempPixelData==NULL) {
		fclose(in);
		return BMP_LOAD_ERROR;
	}
	if (fread(tempPixelData,sizeof(BYTE),size,in) != size)
	{
		fclose(in);
		delete[] tempPixelData;
		return BMP_LOAD_ERROR;
	};
	fclose(in);

	//byteWidth is the width of the actual image in bytes
	//padWidth is the width of the image plus the extra padding
	LONG byteWidth,padWidth;

	//initially set both to the width of the image
	byteWidth=padWidth=(LONG)((float)width*(float)bpp/8.0);

	//add any extra space to bring each line to a DWORD boundary
	while(padWidth%4!=0) {
		padWidth++;
	}
	fclose(in);

	bm.bmBits = tempPixelData;
	bm.bmType = 0;
	bm.bmBitsPixel = bpp;
	bm.bmHeight = height;
	bm.bmWidth = byteWidth;
	bm.bmWidthBytes = padWidth; 
	bm.bmPlanes = bmih.biPlanes;

	HDC dc;
	HBITMAP hbmp;
	CBitmap dummy;

	bi->bmiHeader = bmih;
	dc = GetDC(NULL);
	hbmp = CreateCompatibleBitmap(dc, width, height);
	SetDIBits(dc, hbmp, 0, height, tempPixelData, bi, DIB_RGB_COLORS);
	ReleaseDC(NULL, dc);

	*bmp = dummy.FromHandle(hbmp);

	delete tempPixelData;
	delete bi;

	return ret_val;
}
