#include "Camera.h"
#include "Assert.h"
#include "Application.h"
#include "QuadTreeRenderer.h"

#include "math.h"
#include "stdio.h"

#define pi 3.1415926535
#define MOUSE_DIST 5.0f

unsigned char KEY_STRAFE_LEFT    = 0;
unsigned char KEY_STRAFE_RIGHT   = 0;
unsigned char KEY_MOVE_FORWARD   = 0;
unsigned char KEY_MOVE_BACKWARD  = 0;
unsigned char KEY_MOVE_UP        = 0;
unsigned char KEY_MOVE_DOWN      = 0;
unsigned char KEY_PITCH_UP       = 0;
unsigned char KEY_PITCH_DOWN     = 0;

unsigned char KEY_SHADE_SWITCH   = 0;
unsigned char KEY_POLY_SWITCH    = 0;
unsigned char KEY_TEXTURE_SWITCH = 0;
unsigned char KEY_CULL_SWITCH    = 0;

unsigned char KEY_DETAIL_UP        = 0;
unsigned char KEY_DETAIL_DOWN      = 0;
unsigned char KEY_LOCALDETAIL_UP   = 0;
unsigned char KEY_LOCALDETAIL_DOWN = 0;

Camera::Camera()
{
  renderer = NULL;
  map      = NULL;
  app      = NULL;

  text_switch = true;
}

GEZERROR Camera::Initialize()
{
	/* initialize fps counter */
	ASSERT(app != NULL, "app is NULL!");

	fps = 0;
	fSecsPerFrame = 1;
	QueryPerformanceFrequency((LARGE_INTEGER *)&pf_frequency);
	QueryPerformanceCounter((LARGE_INTEGER *)&fps_start);
	frames = 0;

	/* initialize vectors*/
	vVelocity.x = vVelocity.y = vVelocity.z = 
	vAngularVelocity.x = vAngularVelocity.y = vAngularVelocity.z = 0.0f;

	ASSERT(map != NULL, "map mss ma da scho haben");

	vPos.x = map->width / 2.0f;
	vPos.y = 50;
	vPos.z = 50;

	vRot.x = 0.0f;
	vRot.y = 180.0f;
	vRot.z = 0.0f;

	/* up-vector, static by now */
	vUp.x = 0;
	vUp.y = 1;
	vUp.z = 0;

	vLook.x = 0; 
	vLook.y = 0;
	vLook.z = -1;

	/* pitchAxis is cross product of up and orientation vector */
	vectorCrossProd(&vUp, &vLook, &vPitchAxis);

	fSpeed    = ((Application *)(app))->prefs.GET_FLOAT_PARAM(PARAM_TRANS_SPEED);
	free_move = ((Application *)(app))->prefs.GET_INT_PARAM(PARAM_FREE_MOVE) != 0;

	benchmark = ((Application *)(app))->prefs.GET_INT_PARAM(PARAM_BENCHMARK) != 0;

	shade_switch = true;
	pm_switch = false;
	text_switch = true;


	/* get controls */
	KEY_STRAFE_LEFT    = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_STRAFE_LEFT)[0];
	KEY_STRAFE_RIGHT   = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_STRAFE_RIGHT)[0];
	KEY_MOVE_FORWARD   = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_MOVE_FORWARD)[0];
	KEY_MOVE_BACKWARD  = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_MOVE_BACKWARD)[0];
	KEY_MOVE_UP        = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_MOVE_UP)[0];
	KEY_MOVE_DOWN      = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_MOVE_DOWN)[0];

	KEY_SHADE_SWITCH   = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_SHADE_SWITCH)[0];
	KEY_POLY_SWITCH    = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_POLY_SWITCH)[0];
	KEY_TEXTURE_SWITCH = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_TEXTURE_SWITCH)[0];
	KEY_CULL_SWITCH    = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_CULL_SWITCH)[0];

	KEY_DETAIL_UP        = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_DETAIL_UP)[0];
	KEY_DETAIL_DOWN      = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_DETAIL_DOWN)[0];
	KEY_LOCALDETAIL_UP   = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_LOCALDETAIL_UP)[0];
	KEY_LOCALDETAIL_DOWN = ((Application *)(app))->prefs.GET_STRING_PARAM(PARAM_LOCALDETAIL_DOWN)[0];

	return GEZ_OK;
}

GEZERROR Camera::SetRenderer(Renderer *new_renderer)
{
  ASSERT(new_renderer != NULL, "camera renderer set to NULL");
  renderer = new_renderer;

  return GEZ_OK;
}

Renderer* Camera::GetRenderer()
{
  return renderer;
}

Camera::~Camera()
{
  if (map != NULL)
  {
    delete map;
    map = NULL;
  }
}

void Camera::SetApp(void *application)
{
  app = application;
}

GEZERROR Camera::Render()
{
	Application *appl;
	Vector3f    newvelocity, newangularvelocity;
	__int64 time;

	ASSERT(map != NULL, "setting renderer without loaded map!");

	/* camera movement */
	ASSERT(app != NULL, "app is null");
	appl = (Application *) app;

	Vector3f vT;
	Vector3f vR;

	if (!benchmark)
	{
		if ( appl->bKey[KEY_DETAIL_UP] )
		{
			renderer->IncreaseDetailLevel();
			appl->bKey[KEY_DETAIL_UP] = 0;
		}

		if ( appl->bKey[KEY_DETAIL_DOWN] )
		{
			renderer->DecreaseDetailLevel();
			appl->bKey[KEY_DETAIL_DOWN] = 0;
		}

		if ( appl->bKey[KEY_LOCALDETAIL_UP  ] )
		{
			renderer->IncreaseLocalDetailLevel();
			appl->bKey[KEY_LOCALDETAIL_UP  ] = 0;
		}

		if ( appl->bKey[KEY_LOCALDETAIL_DOWN] )
		{
			renderer->DecreaseLocalDetailLevel();
			appl->bKey[KEY_LOCALDETAIL_DOWN] = 0;
		}

		/* FIXME */
		if ( appl->bKey[KEY_SHADE_SWITCH] )
		{
			if (shade_switch = !shade_switch)
			{
				glShadeModel(GL_SMOOTH);
				g_console->ScrOutput("ShadeModel: smooth\n");
			}
			else
			{
				glShadeModel(GL_FLAT);
				g_console->ScrOutput("ShadeModel: flat\n");
			}

			appl->bKey[KEY_SHADE_SWITCH] = 0;
		}

		if ( appl->bKey[KEY_TEXTURE_SWITCH] )
		{
			if (text_switch = !text_switch)
			{
				glEnable(GL_TEXTURE_2D);
				renderer->texture = true;
				g_console->ScrOutput("texture enabled\n");
			}
			else
			{
				glDisable(GL_TEXTURE_2D);
				renderer->texture = false;
				g_console->ScrOutput("texture disabled\n");
			}

			appl->bKey[KEY_TEXTURE_SWITCH] = 0;
		}

		if ( appl->bKey[KEY_CULL_SWITCH] )
		{
			if (cull_switch = !cull_switch)
			{
				((QuadTreeRenderer*)renderer)->frustum_culling = true;
				g_console->ScrOutput("frustum culling enabled\n");
			}
			else
			{
				((QuadTreeRenderer*)renderer)->frustum_culling = false;
				g_console->ScrOutput("frustum culling disabled\n");
			}

			appl->bKey[KEY_CULL_SWITCH] = 0;
		}


		if ( appl->bKey[KEY_POLY_SWITCH] )
		{
			if (((QuadTreeRenderer*)renderer)->compiled_array)
			{
				g_console->ScrOutput("wireframe not available when using compiled_vertex_array_EXT\n");
				// comp array crashes when switching to GL_LINE
				// TODO: find out why
			}
			else if (pm_switch = !pm_switch)
			{
				glPolygonMode(GL_FRONT, GL_LINE);
				renderer->texture = false;
				g_console->ScrOutput("wireframe mode selected: glPolygonMode(GL_FRONT, GL_LINE)\n");
			}
			else
			{
				glPolygonMode(GL_FRONT, GL_FILL);
				renderer->texture = text_switch;
				g_console->ScrOutput("filled polygon mode selected: glPolygonMode(GL_FRONT, GL_FILL)\n");
			}

			appl->bKey[KEY_POLY_SWITCH] = 0;
		}



		float cosTheta;
		float sinTheta;


		/* TODO: create lookup tables */
		cosTheta = (float)cos((vRot.y / 180.0) * pi);
		sinTheta = (float)sin((vRot.y / 180.0) * pi);

		/* rotate around Y axis */
		vLook.x = sinTheta;
		vLook.z = -cosTheta;
		vLook.y = 0;

		cosTheta = (float)cos(((vRot.y - 90) / 180.0) * pi); /* FIXME */
		sinTheta = (float)sin(((vRot.y - 90) / 180.0) * pi);

		vPitchAxis.x = sinTheta;
		vPitchAxis.z = -cosTheta;
		vPitchAxis.y = 0;

		/* calculate movement */
		vPitchAxis.x *= fSpeed;
		vPitchAxis.z *= fSpeed;
		vLook.x *= fSpeed;
		vLook.z *= fSpeed;

		if( appl->bKey[KEY_STRAFE_LEFT] )
		{
			vT.x += vPitchAxis.x; // Slide Left
			vT.z += vPitchAxis.z; // Slide Left
		}

		if( appl->bKey[KEY_STRAFE_RIGHT] )
		{
			vT.x -= vPitchAxis.x; // Slide Left
			vT.z -= vPitchAxis.z; // Slide Left
		}

		if (free_move)
		{
			if( appl->bKey[KEY_MOVE_UP] )   vT.y += fSpeed; // Slide Down
			if( appl->bKey[KEY_MOVE_DOWN] ) vT.y -= fSpeed; // Slide Up
		}

		if( appl->bKey[KEY_MOVE_FORWARD] )
		{
			vT.x += vLook.x;
			vT.z += vLook.z; // Move Forward
		}
		  
		if( appl->bKey[KEY_MOVE_BACKWARD] )
		{
			vT.x -= vLook.x;
			vT.z -= vLook.z; // Move Backward
		}

		newvelocity        = (vVelocity * 0.9f) + (vT * 0.1f);
		vVelocity          = newvelocity;

		newangularvelocity   = (vAngularVelocity * 0.9f) + (vR * 0.1f);
		vAngularVelocity     = newangularvelocity;

		vT = newvelocity * fSecsPerFrame * 10.0f;
		vR = vAngularVelocity * fSecsPerFrame * 100.0f;

		vPos = vPos + vT;

		if (!free_move)
		{
			float y0, y1, y2, y3;
			float dx, dz;

			// limit movement to map
			if (vPos.x < 0)
			{
				vPos.x = 0;
			}
			if (vPos.z < 0)
			{
				vPos.z = 0;
			}
			if (vPos.x > map->width - 1)
			{
				vPos.x = map->width - 1;
			}
			if (vPos.z > map->height - 1)
			{
				vPos.z = map->height - 1;
			}

			// interpolate 
			dx = vPos.x - (int)vPos.x;
			dz = vPos.z - (int)vPos.z;

			y0 = ((QuadTreeRenderer *)(renderer))->vertex_data[map->width * (int)vPos.z + (int)vPos.x][1];

			if (vPos.x + 1 < map->width)
				y1 = ((QuadTreeRenderer *)(renderer))->vertex_data[map->width * (int)vPos.z + (int)vPos.x + 1][1];
			else
				y1 = y0;

			y1 = y0 * (1 - dx) + (y1 * (dx));

			if (vPos.z + 1 < map->height)
			{
				y2 = ((QuadTreeRenderer *)(renderer))->vertex_data[map->width * ((int)vPos.z + 1) + (int)vPos.x][1];
				if (vPos.x + 1 < map->width)
					y3 = ((QuadTreeRenderer *)(renderer))->vertex_data[map->width * ((int)vPos.z + 1) + (int)vPos.x + 1][1];
				else
					y3 = y2;
			}
			else
			{
				y2 = y0;
				y3 = y0;
			}

			y2 = y2 * (1 - dx) + (y3 * (dx));

			y0 = y1 * (1 - dz) + (y2 * (dz));

			vPos.y = y0 + 3;
		}

		vRot = vRot + vR;

		/* check for full turn */
		if (vRot.x < 0.0f) vRot.x = 360.0f + vRot.x;
		if (vRot.x >= 360.0f) vRot.x = vRot.x - 360.0f;
		if (vRot.y < 0.0f) vRot.y = 360.0f + vRot.y;
		if (vRot.y >= 360.0f) vRot.y = vRot.y - 360.0f;
		if (vRot.z < 0.0f) vRot.z = 360.0f + vRot.z;
		if (vRot.z >= 360.0f) vRot.z = vRot.z - 360.0f;

	}

	if (renderer != NULL)
	renderer->Render(vPos, vRot);

	if (!benchmark)
	{
		/* calculate frames per second */  
		QueryPerformanceCounter((LARGE_INTEGER *)&time);

		fps = (++frames) / ((float)(time - fps_start) / (float)pf_frequency);
		fSecsPerFrame = (float)1 / float(fps);

		if (time > fps_start + pf_frequency)
		{
			// one second passed
			QueryPerformanceCounter((LARGE_INTEGER *)&fps_start);
			frames = 0;
		}
	}
	else
	{
		vPos.z += 0.1;
	}


	return GEZ_OK;
}

GEZERROR Camera::LoadMap(char *fname)
{
  ASSERT(map == NULL, "map is not null");
  
  map = new Map();
  if (map == NULL)
  {
    return GEZ_UNKNOWN;
  }

  return map->LoadMap(fname);
}

void Camera::RotateCam(float yaw, float pitch)
{
	vRot.y += yaw /   MOUSE_DIST;
	vRot.x += pitch / MOUSE_DIST;

	if (vRot.x < 0.0f) vRot.x = 360.0f + vRot.x;
	if (vRot.x >= 360.0f) vRot.x = vRot.x - 360.0f;
	if (vRot.y < 0.0f) vRot.y = 360.0f + vRot.y;
	if (vRot.y >= 360.0f) vRot.y = vRot.y - 360.0f;
}

