#include "preferences.h"
#include "Error.h"
#include "Assert.h"
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifndef NULL
#define NULL 0x0L
#endif


Preferences::Preferences()
{
	sym_table  = NULL;
	pref_names = NULL;
	pref_count = 0;
}

Preferences::~Preferences()
{
	if (pref_names)
	{
		delete pref_names;
		pref_names = NULL;
	}

	if (sym_table)
	{
		delete sym_table;
		sym_table = NULL;
	}
}

int Preferences::LookUp(char *name)
{
	int i;
	bool found;

	found = false;
	i = 0;

	while (i < pref_count)
	{
		if (strcmp(pref_names[i], name) == 0)
		{
			found = true;
			break;
		}

		++i;
	}

	if (found)
	{
		return i;
	}
	else
		return PARAM_ERROR;
}

GEZERROR Preferences::putsym(char *sym_name, int type, sym_data value)
{
    int i;
    char *dummy;

	if ((i = LookUp(sym_name)) != PARAM_ERROR)
	{
		if (sym_table[i].type == type)
		{
			switch (type)
			{
				case PARAM_INT:
					sym_table[i].data.int_val = value.int_val;
					break;
				case PARAM_FLOAT:
					sym_table[i].data.float_val = value.float_val;
					break;
				case PARAM_STRING:

					/* FIXME leaking, but default string cannot be deleted */
/*					if (sym_table[i].data.string_val != NULL)
					{
						delete sym_table[i].data.string_val;
					}
*/
					/* FIXME memory is leaked... */
					dummy = new char[strlen(value.string_val) + 1];
					ASSERT(dummy != NULL, "out of memory, eh?");
					strcpy(dummy, value.string_val);

					sym_table[i].data.string_val = dummy;
					break;
			}
			return GEZ_OK;
		}
	}

	return PARAM_ERROR;
}

GEZERROR Preferences::create_sym(PrefNames pref, int type, sym_data value)
{
    char *dummy;

	if (pref < pref_count)
	{
		sym_table[pref].type = type;

		switch (type)
		{
			case PARAM_INT:
				sym_table[pref].data.int_val = value.int_val;
				break;
			case PARAM_FLOAT:
				sym_table[pref].data.float_val = value.float_val;
				break;
			case PARAM_STRING:
				/* FIXME memory is leaked... */
				dummy = new char[strlen(value.string_val) + 1];
				ASSERT(dummy != NULL, "out of memory, eh?");
				strcpy(dummy, value.string_val);

				sym_table[pref].data.string_val = value.string_val;
				break;
		}

		return GEZ_OK;
	}

	return PARAM_ERROR;
}

GEZERROR Preferences::putsym(PrefNames pref, int type, sym_data value)
{
	char *dummy;

	if (pref < pref_count)
	{
		if (sym_table[pref].type == type)
		{
			switch (type)
			{
				case PARAM_INT:
					sym_table[pref].data.int_val = value.int_val;
					break;
				case PARAM_FLOAT:
					sym_table[pref].data.float_val = value.float_val;
					break;
				case PARAM_STRING:
					/* FIXME memory is leaked... */
					dummy = new char[strlen(value.string_val) + 1];
					ASSERT(dummy != NULL, "out of memory, eh?");
					strcpy(dummy, value.string_val);

					sym_table[pref].data.string_val = value.string_val;
					break;
			}

		}

		return GEZ_OK;
	}

	return PARAM_ERROR;
}


sym_data Preferences::getsym(PrefNames pref, int type)
{
	if (pref < pref_count)
	{
		return sym_table[pref].data;
	}

	return PARAM_ERROR;
}

int Preferences::getsymtype(char *sym_name)
{
	int i;

	if ((i = LookUp(sym_name)) != PARAM_ERROR)
	{
		return sym_table[i].type;
	}

	return PARAM_ERROR;
}

bool Preferences::LoadConfig(char *fname)
{
	FILE *file;
	char buffer[64];
	int i, j, k, type;

    /* insert default parameters into sym_table */
    #ifdef PARAMETER
    #undef PARAMETER
    #endif

    /* first, find out how many there are */
	#define PARAMETER(name, type, value, dummy) ++pref_count;

	pref_count = 0;
    #include "params.h"
	#undef PARAMETER

	/* allocate preferences name table and sym table */
	pref_names = new char*[pref_count];
	sym_table   = new symtable_tag[pref_count];

	memset(sym_table, 0, pref_count * sizeof(symtable_tag));

    i = 0;

	/* FIXME: error check */

    #define QUOTE(t) #t
	#define PARAMETER(name, type, value, dummy) pref_names[i] = QUOTE(name), ++i;
	#include "params.h"
	#undef PARAMETER

	/* insert default values */
    #define PARAMETER(name, type, value, dummy) create_sym(PARAM_ ## name, PARAM_ ## type, (sym_data)value); //knnt ma auch mim int machen
	#include "params.h"
	#undef PARAMETER

	file = fopen(fname, "r");

	if (file == NULL)
	{
		return false;
	}

	while (fgets(buffer, 64, file))
	{
        // strip whitespaces
		i = 0;
		k = 0;
		while (buffer[i] != '\0')
		{
			/* if first character is '#' line is comment -> skip */ 
			if (i == 0 && buffer[i] == '#')
			{
				break;
			}

			if (buffer[i] == ' ' || buffer[i] == '\n')
			{
				for (j = i--; buffer[j] != '\0'; ++j)
				{
					buffer[j] = buffer[j + 1];
				}
			}

			
            if (buffer[i] == '=')
			{
				buffer[i] = '\0';
				k = i + 1;
			}

			i++;
		}


		if (k != 0 && buffer[0] != '#')
		{
			sym_data val;

			switch (type = getsymtype(buffer))
			{
				case PARAM_INT:
					val.int_val = atoi(&(buffer[k]));
					break;
				case PARAM_FLOAT:
					val.float_val = atof(&(buffer[k]));
					break;
				case PARAM_STRING:
					val.string_val = &(buffer[k]);
					break;
				default:
					WARN("unknown parameter type encountered");
					break;
			}

			if (type != PARAM_ERROR)
			{
				GEZERROR result;

				result = putsym(buffer, type, val);

				if (result == PARAM_ERROR) WARN("error inserting parameter");
			}
		}
	}

    fclose(file);

	return true;
}

